import pandas as pd
from antm.ctfidf import CTFIDFVectorizer
from sklearn.feature_extraction.text import CountVectorizer

def rep_prep(cluster_df):
    clusters_df = pd.concat(cluster_df)

    clusters_df_copy=clusters_df.copy()
    clusters_df_copy.loc[:,"num_doc"]=1
    clusters_df=clusters_df_copy

    documents_per_topic_per_time= clusters_df.groupby(["slice_num","C"], as_index=False).agg({'content': ' '.join,"num_doc":"count"})
    documents_per_topic_per_time=documents_per_topic_per_time.reset_index().rename(columns={"index":"cluster"})

    return documents_per_topic_per_time


def ctf_idf_topics(docs_per_class,words,ctfidf,num_terms):
    topics=[]
    for label in docs_per_class:
        topic=[]
        for index in ctfidf[int(label)].argsort()[-num_terms:]:
            topic.append(words[index])
        topics.append(topic)
    return topics

def ctfidf_rp(dictionary,documents_per_topic_per_time,num_doc,num_words=10):
    count_vectorizer= CountVectorizer(vocabulary=dictionary.token2id).fit(documents_per_topic_per_time.content)
    words= count_vectorizer.get_feature_names_out()
    count= count_vectorizer.transform(documents_per_topic_per_time.content)
    ctfidf= CTFIDFVectorizer().fit_transform(count, n_samples=num_doc).toarray()
    topics_representations=ctf_idf_topics(documents_per_topic_per_time.cluster,words,ctfidf,num_words)
    output = documents_per_topic_per_time.assign(topic_representation=topics_representations)
    return output


def topic_evolution(list_tm,output):
    evolving_topics = []
    for et in list_tm:
        evolving_topic = []
        for topic in et:
            cl = int(float(topic.split("-")[1]))
            win = int(float(topic.split("-")[0]))
            t = output[output["slice_num"] == win]
            t = t[t["C"] == cl]
            evolving_topic.append(t.topic_representation.to_list()[0])
        evolving_topics.append(evolving_topic)
    evolving_topics_df = pd.DataFrame({'evolving_topics': evolving_topics})
    return evolving_topics_df



import glob
import os
import pickle
import random
import torch
import pandas as pd
from matplotlib import pyplot as plt
from antm.text_processing import text_processing
from antm.aligned_clustering_layer import aligned_umap, hdbscan_cluster, draw_cluster, clustered_df, plot_alignment, \
    alignment_procedure, dt_creator, clustered_cent_df
from antm.sws import sws
from antm.contextual_embedding_layer import contextual_embedding
from antm.topic_representation_layer import rep_prep,ctfidf_rp, topic_evolution
from antm.cm import coherence_model
from antm.diversity_metrics import proportion_unique_words,pairwise_jaccard_diversity

class ANTM:
    def __init__(self, df, overlap, window_length, mode="bert", umap_dimension_size=5, umap_n_neighbors=15,
                 embedding_vectors=None, umap_embeddings_clustering=None, umap_embeddings_visulization=None,
                 device=None , partioned_clusttering_size=10, num_words=10, show_2d_plot=False, show_3d_plot=False,path=os.getcwd()):
        self.df = df
        self.overlap = overlap
        self.window_length = window_length
        self.mode = mode
        self.umap_dimension_size = umap_dimension_size
        self.umap_n_neighbors = umap_n_neighbors
        self.partioned_clusttering_size = partioned_clusttering_size
        self.num_words = num_words
        self.show_2d_plot = show_2d_plot

        if not path== os.getcwd():
            if not os.path.exists(path): os.mkdir(path)
        self.path = path


        if device==None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device=device

        self.embedding_vectors = embedding_vectors
        self.umap_embeddings_clustering = umap_embeddings_clustering
        self.umap_embeddings_visulization = umap_embeddings_visulization
        self.df_embedded = None
        self.clusters=None
        self.slices=None
        self.arg1_umap=None
        self.arg2_umap=None
        self.cluster_df=None
        self.clustered_df_cent=None
        self.clustered_np_cent=None
        self.dt=None
        self.concat_cent=None
        self.df_tm=None
        self.list_tm=None
        self.documents_per_topic_per_time=None
        self.tokens=None
        self.dictionary=None
        self.corpus=None
        self.output=None
        self.evolving_topics=None
        self.topics=None
        self.slice_num=None
        self.show_3d_plot=show_3d_plot
        self.periodwise_puw_diversity=None
        self.periodwise_pairwise_jaccard_diversity=None
        self.periodwise_topic_coherence=None

    def fit(self, save=True):
        # Contextual embedding
        if self.embedding_vectors is None:
            print("contextual document embedding is initiated...")
            self.df_embedded = contextual_embedding(self.df, mode=self.mode,device=self.device)
        else:
            print("contextual document embedding provided ---> skip")
            self.df["embedded"]=self.embedding_vectors
            self.df_embedded=self.df
        # Sliding window segmentation
        print("Sliding Window Segmentation is initialized...")
        self.slices, self.arg1_umap, self.arg2_umap = sws(self.df_embedded, self.overlap, self.window_length)

        # Aligned dimensionality reduction
        if self.umap_embeddings_clustering is None or self.umap_embeddings_visulization is None:
            print("Aligned Dimension Reduction is initialized...")
            self.umap_embeddings_clustering, self.umap_embeddings_visulization = aligned_umap(
                self.arg1_umap, self.arg2_umap, n_neighbors=self.umap_n_neighbors,
                umap_dimension_size=self.umap_dimension_size)
        else:
            print("umap embeddings provided ---> skip")

        print("Sequential Document-cluster association is initialized...")
        self.clusters = hdbscan_cluster(self.umap_embeddings_clustering,
                                                            self.partioned_clusttering_size)
        if not os.path.exists(self.path + "/results"): os.mkdir(self.path + "/results")
        for i in range(len(self.clusters)):
            draw_cluster(self.clusters[i], self.umap_embeddings_visulization[i], "time_frame_" + str(i),
                         show_2d_plot=self.show_2d_plot, path=self.path)
        self.cluster_df = clustered_df(self.slices, self.clusters)
        self.clustered_df_cent, self.clustered_np_cent = clustered_cent_df(self.cluster_df)
        self.dt, self.concat_cent = dt_creator(self.clustered_df_cent)
        print("Cluster Alignment Procedure is initialized...")
        self.df_tm = alignment_procedure(self.dt, self.concat_cent)
        self.list_tm = plot_alignment(self.df_tm, self.umap_embeddings_visulization, self.clusters, self.path,self.show_3d_plot)
        self.documents_per_topic_per_time = rep_prep(self.cluster_df)
        self.tokens, self.dictionary, self.corpus = text_processing(self.df.content.values)
        print("Topic Representation is initialized...")
        self.output = ctfidf_rp(self.dictionary, self.documents_per_topic_per_time, num_doc=len(self.df),
                                num_words=self.num_words)
        print("Topic Modeling is done")
        self.evolving_topics = topic_evolution(self.list_tm, self.output)
        if save: self.save()
        self.slice_num = set(self.output["slice_num"])
        self.topics = [self.output[self.output["slice_num"] == i].topic_representation.to_list() for i in
                       self.slice_num]
        self.topics = list(filter(None, self.topics))
        return self.topics

    def save(self):
        print("Model is saving...")
        if not os.path.exists(self.path+"/model"): os.mkdir(self.path+"/model")

        self.df_embedded.to_pickle(self.path+"/model/embedding_df")

        with open(self.path+"/model/slices", "wb") as fp:  # Pickling
            pickle.dump(self.slices, fp)

        with open(self.path+"/model/umap_embeddings_clustering", "wb") as fp:  # Pickling
            pickle.dump(self.umap_embeddings_clustering, fp)

        with open(self.path+"/model/umap_embeddings_visulization", "wb") as fp:  # Pickling
            pickle.dump(self.umap_embeddings_visulization, fp)

        with open(self.path+"/model/clusters", "wb") as fp:  # Pickling
            pickle.dump(self.clusters, fp)

        self.df_tm.to_pickle(self.path+"/model/df_tm")

        self.output.to_pickle(self.path + "/model/output")

        self.evolving_topics.to_pickle(self.path + "/model/evolving_topics")


    def load(self):
        print("Model is Loading...")
        self.df_embedded=pd.read_pickle(self.path+"/model/embedding_df")

        with open(self.path + "/model/slices", "rb") as fp:  # Pickling
            self.slices=pickle.load(fp)

        with open(self.path + "/model/umap_embeddings_clustering", "rb") as fp:
            self.umap_embeddings_clustering = pickle.load(fp)

        with open(self.path + "/model/umap_embeddings_visulization", "rb") as fp:
            self.umap_embeddings_visulization = pickle.load(fp)

        with open(self.path + "/model/clusters", "rb") as fp:
            self.clusters = pickle.load(fp)

        self.df_tm=pd.read_pickle(self.path+"/model/df_tm")

        self.list_tm = plot_alignment(self.df_tm, self.umap_embeddings_visulization, self.clusters,self.path,self.show_3d_plot)

        for i in range(len(self.clusters)):
            draw_cluster(self.clusters[i], self.umap_embeddings_visulization[i], "time_frame_" + str(i),
                         show_2d_plot=self.show_2d_plot,path=self.path)

        self.cluster_df = clustered_df(self.slices, self.clusters)

        self.clustered_df_cent, self.clustered_np_cent = clustered_cent_df(self.cluster_df)

        self.dt, self.concat_cent = dt_creator(self.clustered_df_cent)

        self.documents_per_topic_per_time = rep_prep(self.cluster_df)

        self.tokens, self.dictionary, self.corpus = text_processing(self.df.content.values)

        self.output=pd.read_pickle(self.path+"/model/output")

        self.evolving_topics = pd.read_pickle(self.path + "/model/evolving_topics")
        self.slice_num = len(set(self.output["slice_num"]))
        self.topics = [self.output[self.output["slice_num"] == i].topic_representation.to_list() for i in
                    range(1, self.slice_num + 1)]
        return self.topics

    def random_evolution_topic(self):
        random_element = random.choice(self.list_tm)
        list_words = []
        for i in range(len(random_element)):
            # print(random_element[i][0])
            cl = int(float(random_element[i].split("-")[1]))
            win = int(float(random_element[i].split("-")[0]))
            t = self.output[self.output["slice_num"] == win]
            t = t[t["C"] == cl]
            list_words.append(list(t.topic_representation)[0])
            # print(list(t.topic_representation))

        plt.figure(figsize=(15, 10))
        for i in range(len(random_element)):
            cl = int(float(random_element[i].split("-")[1]))
            win = int(float(random_element[i].split("-")[0]))
            labels = self.clusters[win - 1]
            data = self.umap_embeddings_visulization[win - 1]
            data = data.assign(C=labels)
            data = data[data["C"] == cl]
            plt.scatter(data[0], data[1], label=[win,list_words[i]])
            plt.legend()

        plt.savefig(self.path+'/results/random_topic_evolution.png')
        plt.show()

    def save_evolution_topics_plots(self,display=False):
        for j in range(len(self.list_tm)):
            list_words = []
            random_element = self.list_tm[j]
            for i in range(len(random_element)):
                # print(random_element[i])
                cl = int(float(random_element[i].split("-")[1]))
                win = int(float(random_element[i].split("-")[0]))
                t = self.output[self.output["slice_num"] == win]
                t = t[t["C"] == cl]
                list_words.append(list(t.topic_representation)[0])
                # print(list(t.topic_representation))
            fig = plt.figure(figsize=(15, 10))
            for i in range(len(random_element)):
                cl = int(float(random_element[i].split("-")[1]))
                win = int(float(random_element[i].split("-")[0]))
                labels = self.clusters[win - 1]
                data = self.umap_embeddings_visulization[win - 1]
                data = data.assign(C=labels)
                data = data[data["C"] == cl]
                plt.scatter(data[0], data[1], label=[win,list_words[i]])
                plt.legend()
            if not os.path.exists(self.path+"/results/evolving_topics"): os.mkdir(self.path+"/results/evolving_topics")
            plt.savefig(self.path+'/results/evolving_topics/topic_evolution_' + str(j) + '.png')
            if display:
                plt.show()
            plt.close(fig)

    def plot_clusters_over_time(self):

        filenames = glob.glob(self.path+"/results/partioned_clusters/*.png")

        # number of subplots to display
        n_subplots = len(filenames)

        # Plotting figure with 4 rows and 3 columns
        fig, axs = plt.subplots((n_subplots // 4) + 1, 4, figsize=(15, 15))
        axs = axs.ravel()
        ex = len(axs) - n_subplots
        for i in range(ex):
            axs[-i - 1].axis('off')
        # Adding each png file to the figure
        for i, png in enumerate(filenames):
            img = plt.imread(png)
            axs[i].imshow(img)
            axs[i].axis('off')
            axs[i].set_title(f'Time Frame {i + 1}')

        # Adding subtitle to the figure
        plt.suptitle('Dynamic Document Embeddings and Clusters in each Time Frame')
        plt.savefig(self.path + "/results/partioned_topics.png")
        # Showing the figure
        plt.show()

    def plot_evolving_topics(self):

        filenames = glob.glob(self.path+"/results/evolving_topics/*.png")

        # number of subplots to display
        n_subplots = len(filenames)

        # Plotting figure with 3 columns
        fig, axs = plt.subplots((n_subplots // 3) + 1, 3, figsize=(15, 15))
        axs = axs.ravel()
        ex = len(axs) - n_subplots
        for i in range(ex):
            axs[-i - 1].axis('off')
        # Adding each png file to the figure
        for i, png in enumerate(filenames):
            img = plt.imread(png)
            axs[i].imshow(img)
            axs[i].axis('off')
            axs[i].set_title(f'Evolving Topic {i + 1}')

        # Adding subtitle to the figure
        plt.suptitle('Evolution of Evolving Topics')
        # Showing the figure
        plt.savefig(self.path+"/results/evolving_topics.png")
        plt.show()

    def get_periodwise_puw_diversity(self):
        self.periodwise_puw_diversity=[proportion_unique_words(period, topk=self.num_words) for period in self.topics]
        return  self.periodwise_puw_diversity

    def get_periodwise_pairwise_jaccard_diversity(self):
        self.periodwise_pairwise_jaccard_diversity=[pairwise_jaccard_diversity(period, topk=self.num_words) for period in self.topics]
        return  self.periodwise_pairwise_jaccard_diversity

    def get_periodwise_topic_coherence(self,model="c_npmi"):
        self.periodwise_topic_coherence=[coherence_model(period,self.tokens,self.dictionary,self.num_words,c_m=model) for period in self.topics]
        return self.periodwise_topic_coherence



def slice_by_year(list_zip, min_year, max_year, nb_years, overlap_years) :
    if(nb_years < overlap_years) :
        return []
    sliced_list_id = []
    sliced_list_dv = []
    list_tmp_id = []
    list_tmp_dv = []
    curr_year = min_year
    while (curr_year < max_year) :
        period = [curr_year + i for i in range(0,nb_years)]
        for i in range(len(list_zip)) :
            if list_zip[i][0] in period :
                list_tmp_id.append(list_zip[i][1])
                list_tmp_dv.append(list_zip[i][2])
        sliced_list_id.append(list_tmp_id)
        sliced_list_dv.append(list_tmp_dv)
        list_tmp_id = []
        list_tmp_dv = []
        curr_year = curr_year + nb_years - overlap_years
    return sliced_list_dv, sliced_list_id

def slice_df(df, t1, t2, w, o):
    slices = []
    slice_start = t1
    slice_end = t1 + w
    slice_num=0
    while slice_end <= t2+1:
        slice_num+=1
        slice = df[(df['time'] >= slice_start) & (df['time'] < slice_end)]
        slice_copy = slice.copy()
        slice_copy.loc[:, "slice_num"] = slice_num
        slice = slice_copy
        slices.append(slice)
        slice_start += w - o
        slice_end += w -o
    return slices

def relation_periodes(list_ids1, list_ids2) :
    D = dict()
    for i in list_ids1 :
        if i in list_ids2 :
            D[list_ids1.index(i)] = list_ids2.index(i)
    return D

def relations_periodes(list_list_id) :
    L = []
    for i in range(len(list_list_id)-1) :
        L.append(relation_periodes(list_list_id[i],list_list_id[i + 1]))
    return L

def sws(df_embedded,overlap,window_length):
    df_triple = [list(x) for x in zip(df_embedded.time.tolist(), df_embedded.index.tolist(), df_embedded.embedding.tolist())]
    slices=slice_df(df_embedded, df_triple[0][0],  df_triple[len(df_triple)-1][0], window_length,overlap)
    sliced_list_doc_vect, sliced_list_id = slice_by_year(df_triple,df_triple[0][0],
                                                               df_triple[len(df_triple)-1][0],window_length,overlap)
    arg1_umap = sliced_list_doc_vect
    arg2_umap = relations_periodes(sliced_list_id)
    return slices,arg1_umap,arg2_umap


from itertools import combinations



def proportion_unique_words(topics, topk=10):
    """
    compute the proportion of unique words

    Parameters
    ----------
    topics: a list of lists of words
    topk: top k words on which the topic diversity will be computed
    """
    if topk > len(topics[0]):
        raise Exception('Words in topics are less than '+str(topk))
    else:
        unique_words = set()
        for topic in topics:
            unique_words = unique_words.union(set(topic[:topk]))
        puw = len(unique_words) / (topk * len(topics))
        return puw


def pairwise_jaccard_diversity(topics, topk=10):
    '''
    compute the average pairwise jaccard distance between the topics 
  
    Parameters
    ----------
    topics: a list of lists of words
    topk: top k words on which the topic diversity
          will be computed
    
    Returns
    -------
    pjd: average pairwise jaccard distance
    '''
    dist = 0
    count = 0
    for list1, list2 in combinations(topics, 2):
        js = 1 - len(set(list1).intersection(set(list2)))/len(set(list1).union(set(list2)))
        dist = dist + js
        count = count + 1
    return dist/count




import numpy as np
import scipy.sparse as sp
from sklearn.utils import check_array
from sklearn.preprocessing import normalize
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted


class CTFIDFVectorizer(TfidfTransformer):
    def __init__(self, *args, **kwargs):
        super(CTFIDFVectorizer, self).__init__(*args, **kwargs)
        self._idf_diag = None

    def fit(self, X: sp.csr_matrix, n_samples: int):
        """Learn the idf vector (global term weights)

        Parameters
        ----------
        X : sparse matrix of shape n_samples, n_features)
            A matrix of term/token counts.

        """

        # Prepare input
        X = check_array(X, accept_sparse=('csr', 'csc'))
        if not sp.issparse(X):
            X = sp.csr_matrix(X)
        dtype = X.dtype if X.dtype in FLOAT_DTYPES else np.float64

        # Calculate IDF scores
        _, n_features = X.shape
        df = np.squeeze(np.asarray(X.sum(axis=0)))
        avg_nr_samples = int(X.sum(axis=1).mean())
        idf = np.log(avg_nr_samples / df)
        self._idf_diag = sp.diags(idf, offsets=0,
                                  shape=(n_features, n_features),
                                  format='csr',
                                  dtype=dtype)
        return self

    def transform(self, X: sp.csr_matrix, copy=True) -> sp.csr_matrix:
        """Transform a count-based matrix to c-TF-IDF

        Parameters
        ----------
        X : sparse matrix of (n_samples, n_features)
            a matrix of term/token counts

        Returns
        -------
        vectors : sparse matrix of shape (n_samples, n_features)

        """

        # Prepare input
        X = check_array(X, accept_sparse='csr', dtype=FLOAT_DTYPES, copy=copy)
        if not sp.issparse(X):
            X = sp.csr_matrix(X, dtype=np.float64)

        n_samples, n_features = X.shape

        # idf_ being a property, the automatic attributes detection
        # does not work as usual and we need to specify the attribute
        # name:
        check_is_fitted(self, attributes=["idf_"],
                        msg='idf vector is not fitted')

        # Check if expected nr features is found
        expected_n_features = self._idf_diag.shape[0]
        if n_features != expected_n_features:
            raise ValueError("Input has n_features=%d while the model"
                             " has been trained with n_features=%d" % (
                                 n_features, expected_n_features))

        X = X * self._idf_diag

        if self.norm:
            X = normalize(X, axis=1, norm='l1', copy=False)

        return X



from antm.data2vec import data2vec_embedding
from sentence_transformers import SentenceTransformer
from transformers import  Data2VecTextModel, RobertaTokenizer
from transformers import pipeline
import swifter


def contextual_embedding(df,mode,device):
    if mode=="bert":
        df["embedding"] = list(SentenceTransformer('all-mpnet-base-v2',device=device).encode(df.content.to_list(),show_progress_bar=True))
    elif mode=="data2vec":
        tokenizer = RobertaTokenizer.from_pretrained("facebook/data2vec-text-base")
        model = Data2VecTextModel.from_pretrained("facebook/data2vec-text-base").to(device)
        summarizer = pipeline("summarization", model="facebook/bart-large-cnn")
        df["embedding"] = df["content"].swifter.apply(lambda x: data2vec_embedding(x,tokenizer, model, summarizer,device))
    else:
        df["embedding"] = list(SentenceTransformer('all-MiniLM-L6-v2',device=device).encode(df.content.to_list(),show_progress_bar=True))
    return df




from gensim.corpora.dictionary import Dictionary
from gensim.models import CoherenceModel


def coherence_model(topics,tokens,dictionary,tr_num,c_m="c_npmi"):
    cm = CoherenceModel(topics=topics, texts=tokens, dictionary=dictionary,
                                  coherence=c_m, processes=1, topn=tr_num).get_coherence()
    return cm





import os
import umap
import hdbscan
import pandas as pd
import matplotlib.pyplot as plt
import plotly.express as px
import numpy as np


def aligned_umap(arg1_umap,arg2_umap,n_neighbors=15,umap_dimension_size=5):
    model_umap_clustering = umap.aligned_umap.AlignedUMAP(
    metric="cosine",
    n_neighbors=n_neighbors,
    n_components=umap_dimension_size,
    ).fit(arg1_umap, relations = arg2_umap)

    umap_embeddings_clustering=[]
    for j in model_umap_clustering.embeddings_:
        umap_embeddings_clustering.append(pd.DataFrame(j))

    if umap_dimension_size==2:
        umap_embeddings_visulization=umap_embeddings_clustering.copy()
    else:
        model_umap_visualization = umap.aligned_umap.AlignedUMAP(
        metric="cosine",
        n_neighbors=n_neighbors,
        ).fit(arg1_umap, relations = arg2_umap)
        umap_embeddings_visulization=[]
        for j in model_umap_visualization.embeddings_:
            umap_embeddings_visulization.append(pd.DataFrame(j))

    return umap_embeddings_clustering,umap_embeddings_visulization


def hdbscan_cluster(embedding, size) :
    clusters_labels = []
    c= hdbscan.HDBSCAN(min_cluster_size=size, metric = "euclidean",cluster_selection_method = "eom")
    for e in range(len(embedding)) :
        c.fit(embedding[e])
        clusters_labels.append(c.labels_)
    return clusters_labels

def draw_cluster(cluster_labels,umap,name,show_2d_plot,path):
    labels = cluster_labels
    data=umap
    data = data.assign(C=labels)
    data=data[data["C"]>-1]
    fig = plt.figure(figsize=(15, 10))
    plt.scatter(data[0], data[1], c=data["C"])
    if not os.path.exists(path+"/results/partioned_clusters"): os.mkdir(path+"/results/partioned_clusters")
    plt.savefig(path+"/results/partioned_clusters/"+name+'.png')
    if not show_2d_plot:
        plt.close(fig)
    plt.show()

def clustered_df(slices,clusters_labels):
    clustered_df=[]
    for i in range(len(slices)):
        s=slices[i]
        labels=clusters_labels[i]
        s = s.assign(C= labels)
        s=s[s["C"]>-1]
        s=s.reset_index(drop=True)
        clustered_df.append(s)
    return clustered_df


def clustered_cent_df(clustered_df):
    clustered_df_cent=[]
    clustered_np_cent=[]
    for i in clustered_df:
        de=i[["C","embedding"]]
        de = de.groupby("C")["embedding"].apply(list).reset_index()
        de["embedding_mean"] = de["embedding"].apply(lambda x: np.mean(x, axis=0))
        de=pd.DataFrame(list(de['embedding_mean']))
        clustered_df_cent.append(de)
        clustered_np_cent.append(de.to_numpy())
    return clustered_df_cent,clustered_np_cent


def dt_creator(clustered_df_cent):
    topics_cent=[]
    for i in range(len(clustered_df_cent)):
        t=clustered_df_cent[i].copy().reset_index().rename(columns={"index":"cluster_num"})
        for j in range(len(t)):
            t.loc[j, "window_num"] = i+1
        topics_cent.append(t)
    dt=pd.concat(topics_cent).reset_index(drop=True)
    concat_cent=pd.concat(clustered_df_cent).reset_index(drop=True)
    return dt,concat_cent


def alignment_procedure(dt,concat_cent,umap_n_neighbor=5,umap_n_components=2,min_cluster_size=3):
    umap_args = {'n_neighbors': umap_n_neighbor,
                         'n_components': umap_n_components,
                         'metric': 'cosine'}

    hdbscan_args = {'min_cluster_size': min_cluster_size,
                    'metric': 'euclidean',
                    'cluster_selection_method': 'eom'}

    umap_cent = umap.UMAP(**umap_args).fit(concat_cent)
    cluster_cent = hdbscan.HDBSCAN(**hdbscan_args).fit(umap_cent.embedding_)
    return dt.assign(C=cluster_cent.labels_)



def plot_alignment(df_tm,umap_embeddings_visualization,clusters_labels,path,show_3d_plot):
    tm = df_tm[["window_num", "cluster_num", "C"]]
    tm_copy = tm.copy()
    tm_copy.loc[:, "name"] = tm.apply(lambda row: str(row["window_num"]) + "-" + str(row["cluster_num"]) , axis=1)
    tm = tm_copy
    tm = tm[tm["C"] != -1]
    tm = tm.groupby("C")["name"].apply(list).reset_index()
    list_tm = list(tm["name"])
    ccs_list=[]
    for i in range(len(list_tm)):
        cc_list=[]
        for j in list_tm[i]:
            cl=int(float(j.split("-")[1]))
            win=int(float(j.split("-")[0]))
            labels = clusters_labels[win-1]
            data=umap_embeddings_visualization[win-1]
            data = data.assign(C=labels)
            data=data[data["C"]==cl]
            data_copy = data.copy()
            data_copy.loc[:, "win"] = win
            data = data_copy
            cc_list.append(data)
        cc_df=pd.concat(cc_list)
        cc_df_copy=cc_df.copy()
        cc_df_copy.loc[:, "evolving_topic"] = i
        cc_df=cc_df_copy
        ccs_list.append(cc_df)
    ccs_df=pd.concat(ccs_list)

    fig = px.scatter_3d(x=ccs_df[0], y=ccs_df[1], z=ccs_df["win"],
                        color=ccs_df["evolving_topic"],color_continuous_scale=px.colors.sequential.Viridis)
    fig.update_layout(width=1000, height=1000)
    if show_3d_plot:
        fig.show()
    fig.write_image(path+"/results/fig_3D.png")
    return(list_tm)



from setuptools import setup, find_packages

setup(
    name='antm',
    version='0.1.6',
    author='Hamed Rahimi',
    author_email='<hamed.rahimi@sorbonne-universite.fr',
    description='Aligned Neural Topic Model for Exploring Evolving Topics',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/hamedR96/ANTM',
    packages=find_packages(),
    install_requires=["hdbscan","swifter","sentence-transformers","umap-learn","transformers","torch",
                      "gensim","plotly","matplotlib","pyarrow","kaleido"
                      ],
    
    project_urls={
        'Bug Tracker': 'https://github.com/hamedR96/ANTM/issues'
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X ",
        "Operating System :: Unix ",
        "Operating System :: Microsoft :: Windows ",
    ],

)



import torch

def mean_pooling(model_output, attention_mask):
    token_embeddings = model_output[0]
    input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
    embeddings = torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)
    return embeddings[0].detach().cpu().numpy()


def data2vec_embedding(sentence, tokenizer, model,summarizer,device):
    encoded_input = tokenizer(sentence, return_tensors='pt').to(device)
    # Pass the input through the model
    try:
        model_output = model(**encoded_input)
        return mean_pooling(model_output, encoded_input["attention_mask"])
    except:
        print("Summarizing a document with BART due to its Large length for Embedding...")
        new_sentences=summarizer(sentence, max_length=512, do_sample=False)[0]["summary_text"]
        return(data2vec_embedding(new_sentences,tokenizer, model,summarizer,device))




from collections import defaultdict
from nltk import word_tokenize, WordNetLemmatizer
from gensim.corpora import Dictionary
from nltk.corpus import stopwords
import re


def preprocessing_documents(document_list):
    cleaned_documents=[doc.lower() for doc in document_list]
    cleaned_documents = [doc.replace("\n", " ") for doc in cleaned_documents]
    cleaned_documents = [doc.replace("\t", " ") for doc in cleaned_documents]
    cleaned_documents = [re.sub(r'[^A-Za-z0-9 ]+', '', doc) for doc in cleaned_documents]
    return cleaned_documents


def document_tokenize(cleaned_documents):
    stop_words = list(set(stopwords.words('english')))
    document_tokens = [word_tokenize(document) for document in cleaned_documents ]
    tokens=[[token for token in document_tokens if not token in stop_words] for document_tokens in document_tokens ]
    return tokens

def doucments_lemmatizer(documents_tokens):
    tokens=[[WordNetLemmatizer().lemmatize(token) for token in doc_tokens] for doc_tokens in documents_tokens]
    return tokens

def token_frequency_filter(documents_tokens,threshold):
    frequency = defaultdict(int)
    for doc_tokens in documents_tokens:
        for token in doc_tokens:
            frequency[token] += 1
    tokens = [[token for token in doc_tokens if frequency[token] > threshold] for doc_tokens in documents_tokens]
    return tokens


def text_processing(all_documents):
    preprocessed_documents=preprocessing_documents(all_documents)
    documents_tokens=document_tokenize(preprocessed_documents)
    tokens=doucments_lemmatizer(documents_tokens)
    tokens=token_frequency_filter(tokens,5)
    dictionary = Dictionary(tokens)
    corpus = [dictionary.doc2bow(text) for text in tokens]

    return tokens,dictionary,corpus